"""Console layout for mdns-beacon."""
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Tuple, Union

from rich.console import RenderableType
from rich.live import Live
from rich.spinner import Spinner
from rich.table import Table
from rich.text import Text
from zeroconf import IPVersion, ServiceStateChange, Zeroconf


class BaseLayout(ABC):
    """Base cli layout.

    Note:
        Derived layouts must override the `renderable` property.
    """

    _spinner: Optional[Spinner] = None

    def __init__(self, live: Live) -> None:
        """Init layout."""
        self.live = live
        self.live.update(self.renderable)

    @property
    @abstractmethod
    def spinner_text(self) -> str:
        """Get the spinner text to render.

        Property that derived layouts must override.
        """

    @property
    def spinner(self) -> Spinner:
        """Spinner status annimation."""
        if not self._spinner:
            self._spinner = Spinner("dots", text=Text(self.spinner_text, style="green"))
        return self._spinner

    @property
    @abstractmethod
    def renderable(self) -> RenderableType:
        """Get the renderable layout.

        Property that derived layouts must override.
        """


class BlinkLayout(BaseLayout):
    """Blink cli layout."""

    spinner_text = "Announcing services (Press CTRL+C to quit) ..."

    @property
    def renderable(self) -> Table:
        """Blink renderable layout (spinner with status)."""
        layout = Table.grid(padding=1, expand=True)
        layout.add_row(self.spinner)
        return layout


class ListenLayout(BaseLayout):
    """Listen cli layout."""

    services: Dict[str, Any] = {}
    TABLE_SERVICES_COLUMNS = {
        "type": "Type",
        "name": "Name",
        "ipv4_address": "Address IPv4",
        "ipv6_address": "Address IPv6",
        "port": "Port",
        "server": "Server",
        "ttl": "TTL",
        "weight": "Weight",
        "priority": "Priority",
        "text": "TXT",
        "properties": "Properties",
    }
    DEFAULT_SHOW_COLUMNS = (
        "type",
        "name",
        "ipv4_address",
        "port",
        "server",
        "ttl",
    )

    spinner_text = "Listen for services (Press CTRL+C to quit) ..."

    def __init__(
        self,
        show_columns: Optional[Union[Tuple[str], List[str]]] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Init listen layout."""
        self.show_columns = show_columns or self.DEFAULT_SHOW_COLUMNS
        if not set(self.show_columns).issubset(self.TABLE_SERVICES_COLUMNS.keys()):
            raise ValueError(
                "Unknown fields %s", set(self.show_columns) - self.TABLE_SERVICES_COLUMNS.keys()
            )
        super().__init__(*args, **kwargs)

    @property
    def services_table(self) -> Table:
        """Listen services table."""
        table = Table(expand=True)
        table.title = (
            "\n"
            ":police_car_light::satellite_antenna:"
            " mDNS Beacon Listener "
            ":satellite_antenna::police_car_light:"
        )

        table.add_column("#", no_wrap=True)
        for c in self.show_columns:
            table.add_column(self.TABLE_SERVICES_COLUMNS[c], no_wrap=True)

        for index, service in enumerate(self.services.values()):
            table.add_row(str(index), *[str(service[c]) for c in self.show_columns])
        return table

    @property
    def renderable(self) -> RenderableType:
        """Listen renderable layout (a table with spinner and status)."""
        layout = Table.grid(padding=1, expand=True)
        layout.add_row(self.services_table)
        layout.add_row(self.spinner)
        return layout

    def update_services(
        self, zeroconf: Zeroconf, service_type: str, name: str, state_change: ServiceStateChange
    ) -> None:
        """On service state change handler."""
        service_id = f"{name}_{service_type}"
        if state_change is ServiceStateChange.Removed:
            self.services.pop(service_id, None)
        else:
            info = zeroconf.get_service_info(service_type, name)
            if info:
                self.services[service_id] = {
                    "type": info.type,
                    "name": info.name,
                    "ipv4_address": ",".join(info.parsed_addresses(IPVersion.V4Only)),
                    "ipv6_address": ",".join(info.parsed_addresses(IPVersion.V6Only)),
                    "port": info.port,
                    "server": info.server,
                    "ttl": info.host_ttl,
                    "weight": info.weight,
                    "priority": info.priority,
                    "text": info.text.decode("utf8"),
                    "properties": {
                        k.decode("utf8"): v.decode("utf8") for k, v in info.properties.items()
                    },
                }
        self.live.update(self.renderable)



"""Param types for mdns-beacon."""
from ipaddress import IPv4Address, IPv6Address, ip_address
from typing import AnyStr, Optional, Union

import click


class IpAddress(click.ParamType):
    """An IPv4Address or IPv6Address parsed via ipaddress.ip_address.

    Example:
        >>> ptype = IpAddress()
        >>> ptype.convert("127.0.0.1", None, None)
        IPv4Address('127.0.0.1')
    """

    name = "ip_address"

    def convert(
        self, value: AnyStr, param: Optional[click.Parameter], ctx: Optional[click.Context]
    ) -> Optional[Union[IPv4Address, IPv6Address]]:
        """Parse value into IPv4Address or IPv6Address."""
        try:
            return ip_address(value)
        except ValueError:
            self.fail(f"expected an IPv4 or IPv6 address, got {value!r}", param, ctx)



"""Base mDNS Beacon module."""
import asyncio
import logging
from abc import ABC, abstractmethod
from typing import Optional

from zeroconf import IPVersion, Zeroconf

logger = logging.getLogger(__name__)


class BaseBeacon(ABC):
    """mDNS Beacon base class.

    Note:
        Derived beacons must override the `_execute` method.

    Attributes:
        ip_version: IP protocol version to use.
    """

    _zeroconf: Optional[Zeroconf] = None

    def __init__(self, ip_version: Optional[IPVersion] = None) -> None:
        """Init a mDNS Beacon instance.

        Args:
            ip_version: IP protocol version to use.
        """
        self.ip_version = ip_version

    @property
    def zeroconf(self) -> Zeroconf:
        """Zeroconf instance."""
        if not self._zeroconf:
            self._zeroconf = Zeroconf(ip_version=self.ip_version)
        return self._zeroconf

    def stop(self) -> None:
        """Stop Beacon.

        Ends zeroconf background threads, and prevent zeroconf instance from
        servicing further queries.
        """
        logger.debug("Stoping zeroconf")
        self.zeroconf.close()
        self._zeroconf = None

    @abstractmethod
    def _execute(self) -> None:
        """Execute Beacon work.

        Method that derived beacons must override.
        """

    def run_forever(self) -> None:
        """Run beacon forever."""
        self._execute()
        logger.debug("Starting forever loop")
        loop = asyncio.get_event_loop()
        try:
            loop.run_forever()
        finally:
            loop.run_until_complete(loop.shutdown_asyncgens())
            loop.close()



"""mDNS listener module."""
import logging
from typing import Any, Callable, List, Optional, Set, Union

from zeroconf import ServiceBrowser, ServiceListener, ZeroconfServiceTypes

from .base import BaseBeacon

logger = logging.getLogger(__name__)


class BeaconListener(BaseBeacon):
    """mDNS Beacon listener.

    Attributes:
        handlers: Service listeners or functions to be called when a
            service is added, updated or removed.
        services: Fully qualified service type names list.
        timeout: Seconds to wait for any responses.
    """

    _DEFAULT_SERVICES = {"_http._tcp.local.", "_hap._tcp.local."}

    def __init__(
        self,
        handlers: Union[ServiceListener, List[Callable[..., None]]],
        services: Optional[List[str]] = None,
        timeout: Union[int, float] = 5,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Init a mDNS Beacon listener.

        Args:
            handlers: Service listeners or functions to be called when a
                service is added, updated or removed.
            services: Fully qualified service type names list.
            timeout: Seconds to wait for any responses.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        super().__init__(*args, **kwargs)
        self.handlers = handlers
        self.timeout = timeout
        self.services = set(services or self.default_services)

    @property
    def default_services(self) -> Set[str]:
        """Return default services to listen on local networks."""
        return self._DEFAULT_SERVICES | set(
            ZeroconfServiceTypes.find(zc=self.zeroconf, timeout=self.timeout)
        )

    def _execute(self) -> None:
        """Listen for services on the local network."""
        logger.debug("Executing beacon listener")
        ServiceBrowser(zc=self.zeroconf, type_=list(self.services), handlers=self.handlers)



"""Beacon module."""
import logging
import time
from ipaddress import IPv4Address, IPv6Address, ip_address
from typing import Any, Dict, List, Optional, Union

from slugify import slugify
from typing_extensions import Literal
from zeroconf import ServiceInfo

from .base import BaseBeacon

logger = logging.getLogger(__name__)

PROTOCOL = Literal["tcp", "udp"]


class Beacon(BaseBeacon):
    """mDNS Beacon.

    Attributes:
        aliases: Service alias name list.
        addresses: IP addresses that the service runs on.
        port: Port that the service runs on.
        type_: Service type.
        protocol: Service protocol.
        ttl: TTL used for the announce of the service.
        weight: Weight of the service.
        priority: Priority of the service.
        properties: Dict of properties (or a bytes object with the content of the `text` field).
        delay_startup: Amount of time to wait before trying to start
            the zeroconf service (in seconds).
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.
    """

    _SLUG_REGEX_PATTERN = r"[^-a-z0-9_.]+"
    _SLUG_SEPARATOR = "-"
    _services: Optional[List[ServiceInfo]] = None

    def __init__(
        self,
        aliases: Optional[List[str]] = None,
        addresses: Optional[List[Union[IPv4Address, IPv6Address]]] = None,
        port: int = 80,
        type_: str = "http",
        protocol: PROTOCOL = "tcp",
        ttl: int = 60,
        weight: int = 0,
        priority: int = 0,
        properties: Optional[Union[bytes, Dict[str, Any]]] = None,
        delay_startup: int = 0,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Init a mDNS Beacon instance.

        Args:
            aliases: Service alias name list.
            addresses: IP addresses that the service runs on.
            port: Port that the service runs on.
            type_: Service type.
            protocol: Service protocol.
            ttl: TTL used for the announce of the service.
            weight: Weight of the service.
            priority: Priority of the service.
            properties: Dict of properties (or a bytes object with the
                content of the `text` field) of the service.
            delay_startup: Amount of time to wait before trying to start
                the zeroconf service (in seconds).
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        super().__init__(*args, **kwargs)
        self.aliases = set(aliases or [])
        self.addresses = set(addresses or [ip_address("127.0.0.1")])
        self.port = port
        self.type_ = type_
        self.protocol = protocol
        self.ttl = ttl
        self.weight = weight
        self.priority = priority
        self.properties = properties or b""
        self.delay_startup = delay_startup

    def _build_service_host(self, name: str) -> str:
        """Build service host for a given name.

        Args:
            name: Service name.

        Returns:
            Fully qualified service host name.
        """
        slug = slugify(
            name, separator=self._SLUG_SEPARATOR, regex_pattern=self._SLUG_REGEX_PATTERN
        )
        return f"{slug}.local."

    def _build_service_name(self, name: str) -> str:
        """Build service name for a given name.

        Args:
            name: Service name.

        Returns:
            Fully qualified service name.
        """
        return f"{name}.{self.service_type}"

    @property
    def service_type(self) -> str:
        """Beacon service type."""
        return f"_{self.type_}._{self.protocol}.local."

    @property
    def services(self) -> List[ServiceInfo]:
        """Services to register on the local network."""
        if not self._services:
            self._services = [
                ServiceInfo(
                    type_=self.service_type,
                    name=self._build_service_name(alias),
                    parsed_addresses=[str(addr) for addr in self.addresses],
                    port=self.port,
                    host_ttl=self.ttl,
                    weight=self.weight,
                    priority=self.priority,
                    properties=self.properties,
                    server=self._build_service_host(alias),
                )
                for alias in self.aliases
            ]
        return self._services

    def stop(self) -> None:
        """Stop Beacon.

        Unregister all the announced services.
        """
        logger.info("Unregistering %(services_len)s services", services_len=len(self.services))
        for service in self.services:
            logger.debug("Unregistering %(service_name)s", service_name=service.name)
            self.zeroconf.unregister_service(service)
        super().stop()

    def _wait(self) -> None:
        """Wait before trying to start the zeroconf service."""
        time.sleep(self.delay_startup)

    def _execute(self) -> None:
        """Register aliases on the local network."""
        self._wait()
        logger.info("Registering %(services_len)s services", services_len=len(self.services))
        for service in self.services:
            logger.debug("Registering %(service_name)s", service_name=service.name)
            self.zeroconf.register_service(service)



"""Main script for mdns-beacon."""
from ipaddress import IPv4Address, IPv6Address
from typing import Dict, Iterable, Tuple, Union

import click
from click_option_group import MutuallyExclusiveOptionGroup, optgroup
from rich.console import Console
from rich.live import Live

from mdns_beacon import Beacon, BeaconListener, __version__
from mdns_beacon.beacon import PROTOCOL
from mdns_beacon.cli.layouts import BlinkLayout, ListenLayout
from mdns_beacon.cli.types import IpAddress

console = Console()


@click.group()
@click.version_option(version=__version__)
def main() -> None:
    """Simple multicast DNS (mDNS) command line interface utility."""  # noqa:  D401


@main.command()
@click.argument("name")
@click.option(
    "--alias", "aliases", default=[], multiple=True, help="Alias to announce on the local network."
)
@click.option(
    "--address",
    "addresses",
    default=[],
    multiple=True,
    type=IpAddress(),
    help="Address to announce on the local network.",
)
@click.option(
    "--port", "port", default=80, type=int, help="Port to announce on the local network."
)
@click.option("--type", "type_", default="http", type=click.STRING, help="Service type.")
@click.option(
    "--protocol",
    "protocol",
    default="tcp",
    type=click.Choice(choices=("tcp", "udp"), case_sensitive=True),
    help="Service protocol.",
)
@click.option(
    "--weight",
    "weight",
    default=0,
    type=int,
    help="Service weight.",
)
@click.option(
    "--priority",
    "priority",
    default=0,
    type=int,
    help="Service priority.",
)
@optgroup.group(
    "Beacon properties",
    cls=MutuallyExclusiveOptionGroup,
    help="Properties to announce on the local network.",
)
@optgroup.option(
    "--txt",
    "txt",
    type=str,
    callback=lambda ctx, param, value: value.encode("utf8"),
    help="Properties to announce on the local network as `text`.",
    default=b"",
)
@optgroup.option(
    "--property",
    "properties",
    type=click.Tuple([str, str]),
    callback=lambda ctx, param, value: {k.encode("utf8"): v.encode("utf8") for k, v in value},
    help="Properties to announce on the local network (key value).",
    multiple=True,
    default=[],
)
@click.option(
    "--delay-startup",
    "delay_startup",
    default=0,
    type=click.IntRange(0, 300),
    help=(
        "In certain configurations, you may want to wait a given amount of time before "
        "trying to start the mDNS beacon. This is typically found when network interfaces "
        "appear only late during system startup and the interface startup priorities are "
        "configured incorrectly. This setting takes any integer value between 0 and 300 seconds."
    ),
)
def blink(
    name: str,
    aliases: Iterable[str],
    addresses: Iterable[Union[IPv4Address, IPv6Address]],
    port: int,
    type_: str,
    protocol: PROTOCOL,
    weight: int,
    priority: int,
    txt: bytes,
    properties: Dict[str, bytes],
    delay_startup: int,
) -> None:
    """Announce aliases on the local network."""
    with Live(console=console, transient=True, auto_refresh=True) as live:
        BlinkLayout(live=live)
        beacon = Beacon(
            aliases=[name, *aliases],
            addresses=list(addresses),
            port=port,
            type_=type_,
            protocol=protocol,
            weight=weight,
            priority=priority,
            properties=properties or txt,
            delay_startup=delay_startup,
        )
        try:
            beacon.run_forever()
        except KeyboardInterrupt:
            console.print("Shutting down ...")
        finally:
            beacon.stop()


@main.command()
@click.option(
    "--service",
    "services",
    default=[],
    multiple=True,
    help="Service to listen for on the local network.",
)
@click.option(
    "--show",
    "show_columns",
    type=click.Choice(list(ListenLayout.TABLE_SERVICES_COLUMNS.keys()), case_sensitive=True),
    callback=lambda ctx, param, value: tuple({v: None for v in value}.keys()),
    multiple=True,
    default=ListenLayout.DEFAULT_SHOW_COLUMNS,
    help="Service info to show.",
)
def listen(services: Iterable[str], show_columns: Tuple[str]) -> None:
    """Listen for services on the local network."""
    with Live(console=console, transient=True, auto_refresh=True) as live:
        layout = ListenLayout(live=live, show_columns=show_columns)
        listener = BeaconListener(services=list(services), handlers=[layout.update_services])
        try:
            listener.run_forever()
        except KeyboardInterrupt:
            console.print("Shutting down ...")
        finally:
            listener.stop()


if __name__ == "__main__":
    main()  # pragma: no cover



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['mdns_beacon', 'mdns_beacon.cli']

package_data = \
{'': ['*']}

install_requires = \
['click-option-group>=0.5.3,<0.6.0',
 'click>=8.0.1,<9.0.0',
 'python-slugify>=5.0.2,<6.0.0',
 'rich>=10.9.0,<11.0.0',
 'typing-extensions>=3.10.0,<4.0.0',
 'zeroconf>=0.36.2,<0.37.0']

entry_points = \
{'console_scripts': ['mdns-beacon = mdns_beacon.cli.main:main']}

setup_kwargs = {
    'name': 'mdns-beacon',
    'version': '0.4.2',
    'description': 'Multicast DNS (mDNS) Beacon to announce multiple CNAME aliases across your local network.',
    'long_description': '# mDNS Beacon\n\n<div align="center">\n\n[![PyPI - Version](https://img.shields.io/pypi/v/mdns-beacon.svg)](https://pypi.python.org/pypi/mdns-beacon)\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/mdns-beacon.svg)](https://pypi.python.org/pypi/mdns-beacon)\n[![Tests](https://github.com/fedejaure/mdns-beacon/workflows/tests/badge.svg)](https://github.com/fedejaure/mdns-beacon/actions?workflow=tests)\n[![Codecov](https://codecov.io/gh/fedejaure/mdns-beacon/branch/main/graph/badge.svg)](https://codecov.io/gh/fedejaure/mdns-beacon)\n[![Read the Docs](https://readthedocs.org/projects/mdns-beacon/badge/)](https://mdns-beacon.readthedocs.io/)\n[![PyPI - License](https://img.shields.io/pypi/l/mdns-beacon.svg)](https://pypi.python.org/pypi/mdns-beacon)\n\n[![Black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![pre-commit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white)](https://github.com/pre-commit/pre-commit)\n[![Language grade: Python](https://img.shields.io/lgtm/grade/python/g/fedejaure/mdns-beacon.svg?logo=lgtm&logoWidth=18)](https://lgtm.com/projects/g/fedejaure/mdns-beacon/context:python)\n[![Contributor Covenant](https://img.shields.io/badge/Contributor%20Covenant-2.0-4baaaa.svg)](https://www.contributor-covenant.org/version/2/0/code_of_conduct/)\n\n</div>\n\nMulticast DNS (mDNS) Beacon to announce multiple CNAME aliases across your local network. Under development. Use by your own risk❗\n\n\n* GitHub repo: <https://github.com/fedejaure/mdns-beacon.git>\n* Documentation: <https://mdns-beacon.readthedocs.io>\n* Free software: MIT\n\n\n## Features\n\n* ✅ Announce multiple aliases on the local network.\n* ✅ Listening utility to discover services during development.\n* ❌ Configuration file.\n* ❌ Windows support.\n\n## Quickstart\n\nInstall `mdns-beacon` from the [Python Package Index][pypi]:\n\n```\n$ pip install mdns-beacon\n```\n\n#### Usage\n\n```\n$ mdns-beacon --help\nUsage: mdns-beacon [OPTIONS] COMMAND [ARGS]...\n\n  Simple multicast DNS (mDNS) command line interface utility.\n\nOptions:\n  --version  Show the version and exit.\n  --help     Show this message and exit.\n\nCommands:\n  blink   Announce aliases on the local network.\n  listen  Listen for services on the local network.\n```\n\nAnnounce an example service:\n\n```\n$ mdns-beacon blink example --alias sub1.example --address 127.0.0.1 --type http --protocol tcp\n⠋ Announcing services (Press CTRL+C to quit) ...\n```\n\nListen to a specific service type:\n\n```\n$ mdns-beacon listen --service _http._tcp.local.\n                                                                                                            \n                                       🚨📡 mDNS Beacon Listener 📡🚨                                       \n┏━━━┳━━━━━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━┳━━━━━━┳━━━━━━━━━━━━━━━━━━━━━┳━━━━━┓\n┃ # ┃ Type              ┃ Name                           ┃ Address IPv4 ┃ Port ┃ Server              ┃ TTL ┃\n┡━━━╇━━━━━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━╇━━━━━━╇━━━━━━━━━━━━━━━━━━━━━╇━━━━━┩\n│ 0 │ _http._tcp.local. │ example._http._tcp.local.      │ 127.0.0.1    │ 80   │ example.local.      │ 120 │\n│ 1 │ _http._tcp.local. │ sub1.example._http._tcp.local. │ 127.0.0.1    │ 80   │ sub1.example.local. │ 120 │\n└───┴───────────────────┴────────────────────────────────┴──────────────┴──────┴─────────────────────┴─────┘\n                                                                                                            \n⠧ Listen for services (Press CTRL+C to quit) ...\n```\n\n## Credits\n\nThis package was created with [Cookiecutter][cookiecutter] and the [fedejaure/cookiecutter-modern-pypackage][cookiecutter-modern-pypackage] project template.\n\n[cookiecutter]: https://github.com/cookiecutter/cookiecutter\n[cookiecutter-modern-pypackage]: https://github.com/fedejaure/cookiecutter-modern-pypackage\n[pypi]: https://pypi.org/\n',
    'author': 'Federico Jaureguialzo',
    'author_email': 'fedejaure@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/fedejaure/mdns-beacon',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1,<3.10',
}


setup(**setup_kwargs)



"""Top-level package for mdns-beacon."""
from .beacon import Beacon  # noqa: F401
from .listener import BeaconListener  # noqa: F401

__author__ = """Federico Jaureguialzo"""
__email__ = "fedejaure@gmail.com"
__version__ = "0.4.2"
